#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "assemble.h"

#ifdef __cplusplus
extern "C" {
#endif
	long double expl(long double);
	long double logl(long double);
#ifdef __cplusplus
}
#endif

/** initialize the helper structure */
static void init_aux(assemble_aux_t *aa)
{
	int k, n, q;
	long double sum_a[257], b[256], q_c[256], tmp[256], fk2[256];
	aa->l_maf = - 10.0 * log(aa->maf) / log(10.0);
	aa->l1_maf = - 10.0 * log(1.0 - aa->maf) / log(10.0);
	// calculate C^n_k
	aa->fk = (double*)calloc(256, sizeof(double));
	aa->lC = (double*)calloc(512 * 512, sizeof(double));
	aa->coef = (double*)calloc(256*256*64, sizeof(double));
	aa->fk[0] = fk2[0] = 1.0;
	for (n = 1; n != 256; ++n) {
		aa->fk[n] = ((!aa->is_alt)? pow(aa->theta, n) : pow(n+1, -aa->theta)) * (1.0 - aa->eta) + aa->eta;
		fk2[n] = aa->fk[n>>1];
	}
	for (n = 1; n != 512; ++n)
		for (k = 1; k <= n; ++k)
			aa->lC[n<<8|k] = lgamma(n+1) - lgamma(k+1) - lgamma(n-k+1);
	for (q = 1; q != 64; ++q) {
		double e = pow(10.0, -q/10.0);
		double le = log(e);
		double le1 = log(1.0-e);
		for (n = 1; n != 256; ++n) {
			double *coef = aa->coef + (q<<16|n<<8);
			sum_a[n+1] = 0.0;
			for (k = n; k >= 0; --k) { // a_k = \sum_{i=k}^n C^n_k \epsilon^k (1-\epsilon)^{n-k}
				sum_a[k] = sum_a[k+1] + expl(aa->lC[n<<8|k] + k*le + (n-k)*le1);
				b[k] = sum_a[k+1] / sum_a[k];
				if (b[k] > 0.99) b[k] = 0.99;
			}
			for (k = 0; k != n; ++k) // c_k
				q_c[k] = -4.343 * fk2[k] * logl(b[k] / e);
			for (k = 1; k != n; ++k) q_c[k] += q_c[k-1]; // \prod_{i=0}^k c_i
			for (k = 0; k <= n; ++k) { // powl() in 64-bit mode seems broken on my Mac OS X 10.4.9
				tmp[k] = -4.343 * logl(1.0 - expl(fk2[k] * logl(b[k])));
				coef[k] = (k? q_c[k-1] : 0) + tmp[k]; 
			}
			if (q == 1000 && n == 20) { // This never happens. I use these codes to look at values when debugging.
				double ttt[257], *tt;
				tt = ttt + 1;
				tt[0] = fk2[k];
				for (k = 1; k != n; ++k) tt[k] = tt[k-1] + fk2[k];
				for (k = 0; k != n; ++k) {
					fprintf(stderr, "%d\t%Lf\t%Lg\t%Lg\t%g\t%Lg\t%Lg\n", k, fk2[k], expl(aa->lC[n<<8|k] + k*le + (n-k)*le1),
							expl(-coef[k]/4.343)*expl(tt[k]*le), coef[k], expl(logl(b[k])*fk2[k]), b[k]);
				}
				exit(0);
			}
		}
	}
}
assemble_aux_t *new_assemble_aux(int is_alt, float hetero_rate, float theta, float eta, float maf,
								 int max_mm, int max_err, int min_q, int is_single, int is_pair_only)
{
	assemble_aux_t *aa = (assemble_aux_t*)calloc(1, sizeof(assemble_aux_t));
	aa->max_mm = max_mm;
	aa->max_err = max_err;
	aa->maf = maf;
	aa->theta = theta;
	aa->eta = eta;
	aa->min_q = min_q;
	aa->is_alt = is_alt;
	aa->is_single = is_single;
	aa->is_pair_only = is_pair_only;
	aa->q_r = -4.343 * log(2.0 * hetero_rate / (1.0 - hetero_rate));
	init_aux(aa);
	return aa;
}
void delete_assemble_aux(assemble_aux_t *aa)
{
	if (aa == 0) return;
	free(aa->lC); free(aa->fk); free(aa->coef);
	free(aa);
}

